/*
 * Copyright (c) 2011 J. Lewis Muir <jlmuir@imca.aps.anl.gov>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package ch.psi.imagej.cbf;

import java.nio.ByteBuffer;

/**
 * I am a base 64 codec implementing the "base64" encoding in RFC 4648. I do
 * not add line feeds to the encoded data.
 */
public class Base64 {
  private static final String ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                                       + "abcdefghijklmnopqrstuvwxyz"
                                       + "0123456789"
                                       + "+/";
  private static final char PAD = '=';

  public static String encode(byte[] value) {
    StringBuffer result = new StringBuffer(encodedSize(value));
    for (int i = 0; i < value.length; i += 3) {
      int octet1 = value[i] & 0xFF;
      int octet2 = (i + 1 < value.length) ? value[i + 1] & 0xFF : 0;
      int octet3 = (i + 2 < value.length) ? value[i + 2] & 0xFF : 0;
      int group = (octet1 << 16) | (octet2 << 8) | octet3;
      int sextetBitmask = 0x3F;
      int sextet1 = (group >> 18) & sextetBitmask;
      int sextet2 = (group >> 12) & sextetBitmask;
      int sextet3 = (group >> 6) & sextetBitmask;
      int sextet4 = group & sextetBitmask;
      result.append(encode(sextet1));
      result.append(encode(sextet2));
      result.append((i + 1 < value.length) ? encode(sextet3) : PAD);
      result.append((i + 2 < value.length) ? encode(sextet4) : PAD);
    }
    return result.toString();
  }

  private static int encodedSize(byte[] value) {
    if ((value.length % 3) == 0) return value.length * 4 / 3;
    if ((value.length % 3) == 1) return (value.length + 2) * 4 / 3;
    return (value.length + 1) * 4 / 3;
  }

  private static char encode(int sextet) {
    return ALPHABET.charAt(sextet);
  }

  public static byte[] decode(String encoded) {
    ByteBuffer result = ByteBuffer.wrap(new byte[decodedSize(encoded)]);
    for (int i = 0; i < encoded.length(); i += 4) {
      boolean atEnd = i + 4 >= encoded.length();
      int sextet1 = decode(encoded.charAt(i));
      int sextet2 = decode(encoded.charAt(i + 1));
      int sextet3 = (atEnd && encoded.charAt(i + 2) == PAD) ? 0 : decode(encoded.charAt(i + 2));
      int sextet4 = (atEnd && encoded.charAt(i + 3) == PAD) ? 0 : decode(encoded.charAt(i + 3));
      int group = (sextet1 << 18) | (sextet2 << 12) | (sextet3 << 6) | sextet4;
      int octetBitmask = 0xFF;
      byte octet1 = (byte)((group >> 16) & octetBitmask);
      byte octet2 = (byte)((group >> 8) & octetBitmask);
      byte octet3 = (byte)(group & octetBitmask);
      result.put(octet1);
      if (!atEnd || encoded.charAt(i + 2) != PAD) result.put(octet2);
      if (!atEnd || encoded.charAt(i + 3) != PAD) result.put(octet3);
    }
    return result.array();
  }

  private static int decodedSize(String encoded) {
    if ((encoded.length() % 4) != 0) {
      throw new IllegalArgumentException("Encoded string has invalid length");
    }
    String pad = String.valueOf(PAD);
    if (encoded.endsWith(pad + pad)) return (encoded.length() * 3 / 4) - 2;
    if (encoded.endsWith(pad)) return (encoded.length() * 3 / 4) - 1;
    return encoded.length() * 3 / 4;
  }

  private static int decode(char encodedSextet) {
    if (encodedSextet >= 'A' && encodedSextet <= 'Z') return encodedSextet - 'A';
    if (encodedSextet >= 'a' && encodedSextet <= 'z') return encodedSextet - 'a' + 26;
    if (encodedSextet >= '0' && encodedSextet <= '9') return encodedSextet - '0' + 52;
    if (encodedSextet == '+') return 62;
    if (encodedSextet == '/') return 63;
    throw new IllegalArgumentException("Letter not in alphabet: '" + encodedSextet + "'");
  }
}
