/*
 * Copyright (c) 2011 J. Lewis Muir <jlmuir@imca.aps.anl.gov>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package ch.psi.imagej.cbf;

import java.nio.ByteOrder;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.regex.Pattern;

import javax.imageio.metadata.IIOInvalidTreeException;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataNode;

import org.w3c.dom.Node;

/**
 * I am the metadata for a CBF image.
 */
public class CbfMetadata extends IIOMetadata {
  public static final String NATIVE_FORMAT_NAME = "imcacat.jcbf.CbfMetadataFormat_1.0";

  private static final Pattern VALID_STRING = Pattern.compile("(?:\\p{Graph}|\\p{Blank})*");
  private static final Pattern HEADER_DETECTOR_COMMA_DELIM = Pattern.compile("\\s*,\\s*");
  private static final Pattern HEADER_DETECTOR_SN_DELIM = Pattern.compile("\\s+SN:\\s*");

  private String identifier = "###CBF: VERSION 1.5, CBFlib v0.7.8 - SLS/DECTRIS PILATUS detectors";
  private String dataBlockName = "image";

  private String headerConvention = "SLS_1.0";
  private String headerDetector;
  private String headerDetectorName;
  private String headerDetectorSerialNumber;
  private String headerDetectorDescription;
  private Date headerDate;
  private float[] headerPixelSize;
  private String headerSensorType;
  private Float headerSensorThickness;
  private Float headerExposureTime;
  private Float headerExposurePeriod;
  private Float headerTau;
  private Integer headerCountCutoff;
  private Float headerThreshold;
  private String headerGainType;
  private Float headerGainVrf;
  private Integer headerNumExcludedPixels;
  private String headerExcludedPixels;
  private String headerFlatField;
  private String headerTrim;
  private Float headerWavelength;
  private float[] headerEnergyRange;
  private Float headerDetectorDistance;
  private Float headerDetectorVOffset;
  private float[] headerBeamXY;
  private Float headerFlux;
  private Float headerFilterTransmission;
  private Float headerStartAngle;
  private Float headerAngleIncrement;
  private Float headerDetector2Theta;
  private Float headerPolarization;
  private Float headerAlpha;
  private Float headerKappa;
  private Float headerPhi;
  private Float headerChi;
  private String headerOscillationAxis;
  private Integer headerNumOscillations;
  private String headerComment;
  private List<String> headerUnrecognizedLines = new ArrayList<String>();

  private CbfCompression dataCompression;
  private int dataSize;
  private int dataId;
  private CbfElementType dataElementType = CbfElementType.SIGNED_32_BIT_INTEGER;
  private ByteOrder dataElementByteOrder = ByteOrder.LITTLE_ENDIAN;
  private byte[] dataMd5;
  private int dataNumElements;
  private int dataWidth;
  private int dataHeight;
  private int dataPadding = 4095;

  @Override
  public boolean isReadOnly() {
    return true;
  }

  @Override
  public Node getAsTree(String formatName) {
    if (!NATIVE_FORMAT_NAME.equals(formatName)) {
      throw new IllegalArgumentException("Format name: " + formatName);
    }
    Node result = new IIOMetadataNode(NATIVE_FORMAT_NAME);
    if (identifier != null) {
      result.appendChild(createAssociation("identifier", identifier));
    }
    if (dataBlockName != null) {
      result.appendChild(createAssociation("dataBlockName", dataBlockName));
    }

    if (headerConvention != null) {
      result.appendChild(createAssociation("headerConvention", headerConvention));
    }
    if (headerDetector != null) {
      result.appendChild(createAssociation("headerDetector", headerDetector));
    }
    if (headerDate != null) {
      DateFormat format = new SimpleDateFormat("yyyy-MMM-dd'T'HH:mm:ss.SSS", Locale.US);
      result.appendChild(createAssociation("headerDate", format.format(headerDate)));
    }
    if (headerPixelSize != null) {
      result.appendChild(createAssociation("headerPixelSize", String.format(Locale.US,
        "%.0fe-6 m x %.0fe-6 m", headerPixelSize[0] * 1e6f, headerPixelSize[1] * 1e6f)));
    }
    if (headerSensorType != null) {
      result.appendChild(createAssociation("headerSensorType", headerSensorType));
    }
    if (headerSensorThickness != null) {
      result.appendChild(createAssociation("headerSensorThickness",
        String.format(Locale.US, "%.6f m", headerSensorThickness)));
    }
    if (headerExposureTime != null) {
      result.appendChild(createAssociation("headerExposureTime",
        String.format(Locale.US, "%.6f s", headerExposureTime)));
    }
    if (headerExposurePeriod != null) {
      result.appendChild(createAssociation("headerExposurePeriod",
        String.format(Locale.US, "%.6f s", headerExposurePeriod)));
    }
    if (headerTau != null) {
      result.appendChild(createAssociation("headerTau",
        String.format(Locale.US, "%.1fe-09 s", 1e9f * headerTau)));
    }
    if (headerCountCutoff != null) {
      result.appendChild(createAssociation("headerCountCutoff",
        String.format(Locale.US, "%d counts", headerCountCutoff)));
    }
    if (headerThreshold != null) {
      result.appendChild(createAssociation("headerThreshold",
        String.format(Locale.US, "%.0f eV", headerThreshold)));
    }
    if (headerGainType != null) {
      result.appendChild(createAssociation("headerGainType", headerGainType));
    }
    if (headerGainVrf != null) {
      result.appendChild(createAssociation("headerGainVrf",
        String.format(Locale.US, "%.3f", headerGainVrf)));
    }
    if (headerNumExcludedPixels != null) {
      result.appendChild(createAssociation("headerNumExcludedPixels", headerExcludedPixels));
    }
    if (headerExcludedPixels != null) {
      result.appendChild(createAssociation("headerExcludedPixels", headerExcludedPixels));
    }
    if (headerFlatField != null) {
      result.appendChild(createAssociation("headerFlatField", headerFlatField));
    }
    if (headerTrim != null) {
      result.appendChild(createAssociation("headerTrim", headerTrim));
    }
    if (headerWavelength != null) {
      result.appendChild(createAssociation("headerWavelength",
        String.format(Locale.US, "%.4f A", headerWavelength)));
    }
    if (headerEnergyRange != null) {
      result.appendChild(createAssociation("headerEnergyRange",
        String.format(Locale.US, "(%.0f, %.0f) eV", headerEnergyRange[0], headerEnergyRange[1])));
    }
    if (headerDetectorDistance != null) {
      result.appendChild(createAssociation("headerDetectorDistance",
        String.format(Locale.US, "%.5f m", headerDetectorDistance)));
    }
    if (headerDetectorVOffset != null) {
      result.appendChild(createAssociation("headerDetectorVOffset",
        String.format(Locale.US, "%.5f m", headerDetectorVOffset)));
    }
    if (headerBeamXY != null) {
      result.appendChild(createAssociation("headerBeamXY",
        String.format(Locale.US, "(%.2f, %.2f) pixels", headerBeamXY[0], headerBeamXY[1])));
    }
    if (headerFlux != null) {
      result.appendChild(createAssociation("headerFlux",
        String.format(Locale.US, "%.4f ph/s", headerFlux)));
    }
    if (headerFilterTransmission != null) {
      result.appendChild(createAssociation("headerFilterTransmission",
        String.format(Locale.US, "%.4f", headerFilterTransmission)));
    }
    if (headerStartAngle != null) {
      result.appendChild(createAssociation("headerStartAngle",
        String.format(Locale.US, "%.4f deg.", headerStartAngle)));
    }
    if (headerAngleIncrement != null) {
      result.appendChild(createAssociation("headerAngleIncrement",
        String.format(Locale.US, "%.4f deg.", headerAngleIncrement)));
    }
    if (headerDetector2Theta != null) {
      result.appendChild(createAssociation("headerDetector2Theta",
        String.format(Locale.US, "%.4f deg.", headerDetector2Theta)));
    }
    if (headerPolarization != null) {
      result.appendChild(createAssociation("headerPolarization",
        String.format(Locale.US, "%.3f", headerPolarization)));
    }
    if (headerAlpha != null) {
      result.appendChild(createAssociation("headerAlpha",
        String.format(Locale.US, "%.4f deg.", headerAlpha)));
    }
    if (headerKappa != null) {
      result.appendChild(createAssociation("headerKappa",
        String.format(Locale.US, "%.4f deg.", headerKappa)));
    }
    if (headerPhi != null) {
      result.appendChild(createAssociation("headerPhi",
        String.format(Locale.US, "%.4f deg.", headerPhi)));
    }
    if (headerChi != null) {
      result.appendChild(createAssociation("headerChi",
        String.format(Locale.US, "%.4f deg.", headerChi)));
    }
    if (headerOscillationAxis != null) {
      result.appendChild(createAssociation("headerOscillationAxis", headerOscillationAxis));
    }
    if (headerNumOscillations != null) {
      result.appendChild(createAssociation("headerNumOscillations",
        String.format(Locale.US, "%d", headerNumOscillations)));
    }
    if (headerComment != null) {
      result.appendChild(createAssociation("headerComment", headerComment));
    }
    int i = 0;
    for (String each : getHeaderUnrecognizedLines()) {
      result.appendChild(createAssociation("headerUnrecognized" + i, each));
      i++;
    }

    if (dataCompression != null) {
      result.appendChild(createAssociation("dataCompression",
        dataCompression.toContentTypeConversions()));
    }
    result.appendChild(createAssociation("dataSize", Integer.toString(dataSize)));
    result.appendChild(createAssociation("dataId", Integer.toString(dataId)));
    if (dataElementType != null) {
      result.appendChild(createAssociation("dataElementType",
        dataElementType.toXBinaryElementType()));
    }
    if (dataElementByteOrder != null) {
      result
        .appendChild(createAssociation("dataElementByteOrder", dataElementByteOrder.toString()));
    }
    if (dataMd5 != null) {
      result.appendChild(createAssociation("dataMd5", Base64.encode(dataMd5)));
    }
    result.appendChild(createAssociation("dataNumElements", Integer.toString(dataNumElements)));
    result.appendChild(createAssociation("dataWidth", Integer.toString(dataWidth)));
    result.appendChild(createAssociation("dataHeight", Integer.toString(dataHeight)));
    result.appendChild(createAssociation("dataPadding", Integer.toString(dataPadding)));

    return result;
  }

  private IIOMetadataNode createAssociation(String key, String value) {
    IIOMetadataNode result = new IIOMetadataNode("association");
    result.setAttribute("key", key);
    result.setAttribute("value", value);
    return result;
  }

  @Override
  public void mergeTree(String formatName, Node root) throws IIOInvalidTreeException {
    throw new IllegalStateException("Read-only");
  }

  @Override
  public void reset() {
    throw new IllegalStateException("Read-only");
  }

  public String getIdentifier() {
    return identifier;
  }

  public void setIdentifier(String newIdentifier) {
    validate(newIdentifier);
    identifier = newIdentifier;
  }

  public String getDataBlockName() {
    return dataBlockName;
  }

  public void setDataBlockName(String newDataBlockName) {
    validate(newDataBlockName);
    dataBlockName = newDataBlockName;
  }

  public String getHeaderConvention() {
    return headerConvention;
  }

  public void setHeaderConvention(String newHeaderConvention) {
    validate(newHeaderConvention);
    headerConvention = newHeaderConvention;
  }

  public String getHeaderDetector() {
    return headerDetector;
  }

  public void setHeaderDetector(String headerDetector) {
    validate(headerDetector);
    this.headerDetector = headerDetector;
    updateHeaderDetectorParts(headerDetector);
  }

  private void updateHeaderDetectorParts(String headerDetector) {
    headerDetectorName = null;
    headerDetectorSerialNumber = null;
    headerDetectorDescription = null;
    if (headerDetector == null) return;
    String[] parts = HEADER_DETECTOR_COMMA_DELIM.split(headerDetector);
    if (parts.length > 1) {
      headerDetectorName = parts[0];
      headerDetectorSerialNumber = parts[1];
      if (parts.length > 2) headerDetectorDescription = parts[2];
      return;
    }
    parts = HEADER_DETECTOR_SN_DELIM.split(headerDetector);
    if (parts.length > 1) {
      headerDetectorName = parts[0];
      headerDetectorSerialNumber = parts[1];
    }
  }

  public String getHeaderDetectorName() {
    return headerDetectorName;
  }

  public String getHeaderDetectorSerialNumber() {
    return headerDetectorSerialNumber;
  }

  public String getHeaderDetectorDescription() {
    return headerDetectorDescription;
  }

  public Date getHeaderDate() {
    return headerDate;
  }

  public void setHeaderDate(Date headerDate) {
    this.headerDate = headerDate;
  }

  public float[] getHeaderPixelSize() {
    return headerPixelSize;
  }

  public void setHeaderPixelSize(float[] headerPixelSize) {
    this.headerPixelSize = headerPixelSize;
  }

  public String getHeaderSensorType() {
    return headerSensorType;
  }

  public void setHeaderSensorType(String headerSensorType) {
    validate(headerSensorType);
    this.headerSensorType = headerSensorType;
  }

  public Float getHeaderSensorThickness() {
    return headerSensorThickness;
  }

  public void setHeaderSensorThickness(Float headerSensorThickness) {
    this.headerSensorThickness = headerSensorThickness;
  }

  public Float getHeaderExposureTime() {
    return headerExposureTime;
  }

  public void setHeaderExposureTime(Float headerExposureTime) {
    this.headerExposureTime = headerExposureTime;
  }

  public Float getHeaderExposurePeriod() {
    return headerExposurePeriod;
  }

  public void setHeaderExposurePeriod(Float headerExposurePeriod) {
    this.headerExposurePeriod = headerExposurePeriod;
  }

  public Float getHeaderTau() {
    return headerTau;
  }

  public void setHeaderTau(Float headerTau) {
    this.headerTau = headerTau;
  }

  public Integer getHeaderCountCutoff() {
    return headerCountCutoff;
  }

  public void setHeaderCountCutoff(Integer headerCountCutoff) {
    this.headerCountCutoff = headerCountCutoff;
  }

  public Float getHeaderThreshold() {
    return headerThreshold;
  }

  public void setHeaderThreshold(Float headerThreshold) {
    this.headerThreshold = headerThreshold;
  }

  public String getHeaderGainType() {
    return headerGainType;
  }

  public void setHeaderGainType(String headerGainType) {
    validate(headerGainType);
    this.headerGainType = headerGainType;
  }

  public Float getHeaderGainVrf() {
    return headerGainVrf;
  }

  public void setHeaderGainVrf(Float headerGainVrf) {
    this.headerGainVrf = headerGainVrf;
  }

  public Integer getHeaderNumExcludedPixels() {
    return headerNumExcludedPixels;
  }

  public void setHeaderNumExcludedPixels(Integer headerNumExcludedPixels) {
    this.headerNumExcludedPixels = headerNumExcludedPixels;
  }

  public String getHeaderExcludedPixels() {
    return headerExcludedPixels;
  }

  public void setHeaderExcludedPixels(String headerExcludedPixels) {
    validate(headerExcludedPixels);
    this.headerExcludedPixels = headerExcludedPixels;
  }

  public String getHeaderFlatField() {
    return headerFlatField;
  }

  public void setHeaderFlatField(String headerFlatField) {
    validate(headerFlatField);
    this.headerFlatField = headerFlatField;
  }

  public String getHeaderTrim() {
    return headerTrim;
  }

  public void setHeaderTrim(String headerTrim) {
    validate(headerTrim);
    this.headerTrim = headerTrim;
  }

  public Float getHeaderWavelength() {
    return headerWavelength;
  }

  public void setHeaderWavelength(Float headerWavelength) {
    this.headerWavelength = headerWavelength;
  }

  public float[] getHeaderEnergyRange() {
    return headerEnergyRange;
  }

  public void setHeaderEnergyRange(float[] headerEnergyRange) {
    this.headerEnergyRange = headerEnergyRange;
  }

  public Float getHeaderDetectorDistance() {
    return headerDetectorDistance;
  }

  public void setHeaderDetectorDistance(Float headerDetectorDistance) {
    this.headerDetectorDistance = headerDetectorDistance;
  }

  public Float getHeaderDetectorVOffset() {
    return headerDetectorVOffset;
  }

  public void setHeaderDetectorVOffset(Float headerDetectorVOffset) {
    this.headerDetectorVOffset = headerDetectorVOffset;
  }

  public float[] getHeaderBeamXY() {
    return headerBeamXY;
  }

  public void setHeaderBeamXY(float[] headerBeamXY) {
    this.headerBeamXY = headerBeamXY;
  }

  public Float getHeaderFlux() {
    return headerFlux;
  }

  public void setHeaderFlux(Float headerFlux) {
    this.headerFlux = headerFlux;
  }

  public Float getHeaderFilterTransmission() {
    return headerFilterTransmission;
  }

  public void setHeaderFilterTransmission(Float headerFilterTransmission) {
    this.headerFilterTransmission = headerFilterTransmission;
  }

  public Float getHeaderStartAngle() {
    return headerStartAngle;
  }

  public void setHeaderStartAngle(Float headerStartAngle) {
    this.headerStartAngle = headerStartAngle;
  }

  public Float getHeaderAngleIncrement() {
    return headerAngleIncrement;
  }

  public void setHeaderAngleIncrement(Float headerAngleIncrement) {
    this.headerAngleIncrement = headerAngleIncrement;
  }

  public Float getHeaderDetector2Theta() {
    return headerDetector2Theta;
  }

  public void setHeaderDetector2Theta(Float headerDetector2Theta) {
    this.headerDetector2Theta = headerDetector2Theta;
  }

  public Float getHeaderPolarization() {
    return headerPolarization;
  }

  public void setHeaderPolarization(Float headerPolarization) {
    this.headerPolarization = headerPolarization;
  }

  public Float getHeaderAlpha() {
    return headerAlpha;
  }

  public void setHeaderAlpha(Float headerAlpha) {
    this.headerAlpha = headerAlpha;
  }

  public Float getHeaderKappa() {
    return headerKappa;
  }

  public void setHeaderKappa(Float headerKappa) {
    this.headerKappa = headerKappa;
  }

  public Float getHeaderPhi() {
    return headerPhi;
  }

  public void setHeaderPhi(Float headerPhi) {
    this.headerPhi = headerPhi;
  }

  public Float getHeaderChi() {
    return headerChi;
  }

  public void setHeaderChi(Float headerChi) {
    this.headerChi = headerChi;
  }

  public String getHeaderOscillationAxis() {
    return headerOscillationAxis;
  }

  public void setHeaderOscillationAxis(String headerOscillationAxis) {
    validate(headerOscillationAxis);
    this.headerOscillationAxis = headerOscillationAxis;
  }

  public Integer getHeaderNumOscillations() {
    return headerNumOscillations;
  }

  public void setHeaderNumOscillations(Integer headerNumOscillations) {
    this.headerNumOscillations = headerNumOscillations;
  }

  public String getHeaderComment() {
    return headerComment;
  }

  public void setHeaderComment(String headerComment) {
    validate(headerComment);
    this.headerComment = headerComment;
  }

  public void setHeaderDefaults() {
    headerDetector = "PILATUS 6M";
    headerDate = new Date();
    headerPixelSize = new float[] { 172e-6f, 172e-6f };
    headerSensorType = "Silicon";
    headerSensorThickness = 0.000320f;
    headerExposureTime = 1.0f;
    headerExposurePeriod = 1.0f;
    headerTau = 124.0e-09f;
    headerCountCutoff = 1048575;
    headerThreshold = 7439.0f;
    headerGainType = "not implemented";
    headerGainVrf = 9.9f;
    headerWavelength = 1.0f;
    headerEnergyRange = new float[] { 0.0f, 0.0f };
    headerDetectorDistance = 0.2f;
    headerDetectorVOffset = 0.0f;
    headerBeamXY = new float[] { 1231.5f, 1263.5f };
    headerFlux = 0.0f;
    headerFilterTransmission = 0.0f;
    headerStartAngle = 0.0f;
    headerAngleIncrement = 0.0f;
    headerDetector2Theta = 0.0f;
    headerPolarization = 0.990f;
    headerAlpha = 0.0f;
    headerKappa = 0.0f;
    headerPhi = 0.0f;
    headerChi = 0.0f;
    headerOscillationAxis = "X, CW";
    headerNumOscillations = 1;
  }

  public List<String> getHeaderUnrecognizedLines() {
    return Collections.unmodifiableList(headerUnrecognizedLines);
  }

  public void addHeaderUnrecognizedLine(String line) {
    validate(line);
    headerUnrecognizedLines.add(line);
  }

  public CbfCompression getDataCompression() {
    return dataCompression;
  }

  public void setDataCompression(CbfCompression newDataCompression) {
    dataCompression = newDataCompression;
  }

  public int getDataSize() {
    return dataSize;
  }

  public void setDataSize(int newDataSize) {
    dataSize = newDataSize;
  }

  public int getDataId() {
    return dataId;
  }

  public void setDataId(int newDataId) {
    dataId = newDataId;
  }

  public CbfElementType getDataElementType() {
    return dataElementType;
  }

  public void setDataElementType(CbfElementType newDataElementType) {
    dataElementType = newDataElementType;
  }

  public ByteOrder getDataElementByteOrder() {
    return dataElementByteOrder;
  }

  public void setDataElementByteOrder(ByteOrder newDataElementByteOrder) {
    dataElementByteOrder = newDataElementByteOrder;
  }

  public byte[] getDataMd5() {
    return dataMd5;
  }

  public void setDataMd5(byte[] newDataMd5) {
    dataMd5 = newDataMd5;
  }

  public int getDataNumElements() {
    return dataNumElements;
  }

  public void setDataNumElements(int newDataNumElements) {
    dataNumElements = newDataNumElements;
  }

  public int getDataWidth() {
    return dataWidth;
  }

  public void setDataWidth(int newDataWidth) {
    dataWidth = newDataWidth;
  }

  public int getDataHeight() {
    return dataHeight;
  }

  public void setDataHeight(int newDataHeight) {
    dataHeight = newDataHeight;
  }

  public int getDataPadding() {
    return dataPadding;
  }

  public void setDataPadding(int newDataPadding) {
    dataPadding = newDataPadding;
  }

  private void validate(String s) {
    if (s == null) return;
    if (VALID_STRING.matcher(s).matches()) return;
    throw new IllegalArgumentException("String contains an illegal character");
  }
}
