/*
 * Copyright (c) 2011 J. Lewis Muir <jlmuir@imca.aps.anl.gov>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package ch.psi.imagej.cbf;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import javax.imageio.IIOImage;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.event.IIOWriteProgressListener;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageOutputStream;
import javax.imageio.stream.MemoryCacheImageOutputStream;

import org.junit.Test;

import ch.psi.imagej.cbf.CbfCompression;
import ch.psi.imagej.cbf.CbfElementType;
import ch.psi.imagej.cbf.CbfImageReader;
import ch.psi.imagej.cbf.CbfImageWriteParam;
import ch.psi.imagej.cbf.CbfImageWriter;
import ch.psi.imagej.cbf.CbfImageWriterSpi;
import ch.psi.imagej.cbf.CbfMetadata;

/**
 * I test <code>CbfImageWriter</code>.
 */
public class CbfImageWriterTest extends TestUtils {
  @Test
  public void testWriteHeader01() throws IOException {
    ByteArrayOutputStream baStream = new ByteArrayOutputStream();
    ImageOutputStream stream = createOutputStream(baStream);
    CbfMetadata mdata = createMetadata();
    CbfImageWriter writer = createWriter(stream);
    writer.write(null, create1x1Image(mdata), null);
    String expected = "# Detector: PILATUS 6M\r\n"
      + "# 2010-Nov-17T15:52:30.834\r\n"
      + "# Pixel_size 172e-6 m x 172e-6 m\r\n"
      + "# Silicon sensor, thickness 0.000320 m\r\n"
      + "# Exposure_time 1.000000 s\r\n"
      + "# Exposure_period 1.000000 s\r\n"
      + "# Tau = 124.0e-09 s\r\n"
      + "# Count_cutoff 1048575 counts\r\n"
      + "# Threshold_setting 7439 eV\r\n"
      + "# Gain_setting not implemented (vrf = 9.900)\r\n"
      + "# Wavelength 1.0000 A\r\n"
      + "# Energy_range (0, 0) eV\r\n"
      + "# Detector_distance 0.20000 m\r\n"
      + "# Detector_Voffset 0.00000 m\r\n"
      + "# Beam_xy (1231.50, 1263.50) pixels\r\n"
      + "# Flux 0.0000 ph/s\r\n"
      + "# Filter_transmission 0.0000\r\n"
      + "# Start_angle 0.0000 deg.\r\n"
      + "# Angle_increment 0.0000 deg.\r\n"
      + "# Detector_2theta 0.0000 deg.\r\n"
      + "# Polarization 0.990\r\n"
      + "# Alpha 0.0000 deg.\r\n"
      + "# Kappa 0.0000 deg.\r\n"
      + "# Phi 0.0000 deg.\r\n"
      + "# Chi 0.0000 deg.\r\n"
      + "# Oscillation_axis X, CW\r\n"
      + "# N_oscillations 1\r\n";
    String actual = extractHeader(baStream.toByteArray());
    assertEquals(expected, actual);
  }

  /* No gain type. */
  @Test
  public void testWriteHeader02() throws IOException {
    ByteArrayOutputStream baStream = new ByteArrayOutputStream();
    ImageOutputStream stream = createOutputStream(baStream);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setHeaderGainVrf(9.9f);
    CbfImageWriter writer = createWriter(stream);
    writer.write(null, create1x1Image(mdata), null);
    String expected = "# Gain_setting not implemented (vrf = 9.900)\r\n";
    String actual = extractHeader(baStream.toByteArray());
    assertEquals(expected, actual);
  }

  /* Excluded pixels. */
  @Test
  public void testWriteHeader03() throws IOException {
    ByteArrayOutputStream baStream = new ByteArrayOutputStream();
    ImageOutputStream stream = createOutputStream(baStream);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setHeaderExcludedPixels("badpix_mask.tif");
    CbfImageWriter writer = createWriter(stream);
    writer.write(null, create1x1Image(mdata), null);
    String expected = "# Excluded_pixels: badpix_mask.tif\r\n";
    String actual = extractHeader(baStream.toByteArray());
    assertEquals(expected, actual);
  }

  /* Flat field. */
  @Test
  public void testWriteHeader04() throws IOException {
    ByteArrayOutputStream baStream = new ByteArrayOutputStream();
    ImageOutputStream stream = createOutputStream(baStream);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setHeaderFlatField("flat_field.tif");
    CbfImageWriter writer = createWriter(stream);
    writer.write(null, create1x1Image(mdata), null);
    String expected = "# Flat_field: flat_field.tif\r\n";
    String actual = extractHeader(baStream.toByteArray());
    assertEquals(expected, actual);
  }

  /* Trim directory. */
  @Test
  public void testWriteHeader05() throws IOException {
    ByteArrayOutputStream baStream = new ByteArrayOutputStream();
    ImageOutputStream stream = createOutputStream(baStream);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setHeaderTrim("p6m0103_T8p0_vrf_m0p3_100323");
    CbfImageWriter writer = createWriter(stream);
    writer.write(null, create1x1Image(mdata), null);
    String expected = "# Trim_directory: p6m0103_T8p0_vrf_m0p3_100323\r\n";
    String actual = extractHeader(baStream.toByteArray());
    assertEquals(expected, actual);
  }

  /* Unrecognized header fields. */
  @Test
  public void testWriteHeader06() throws IOException {
    ByteArrayOutputStream baStream = new ByteArrayOutputStream();
    ImageOutputStream stream = createOutputStream(baStream);
    CbfMetadata mdata = new CbfMetadata();
    mdata.addHeaderUnrecognizedLine("# Favorite_color: Blue");
    mdata.addHeaderUnrecognizedLine("# Favorite_shape: Box");
    CbfImageWriter writer = createWriter(stream);
    writer.write(null, create1x1Image(mdata), null);
    String expected = "# Favorite_color: Blue\r\n# Favorite_shape: Box\r\n";
    String actual = extractHeader(baStream.toByteArray());
    assertEquals(expected, actual);
  }

  /* Comment. */
  @Test
  public void testWriteHeader07() throws IOException {
    ByteArrayOutputStream baStream = new ByteArrayOutputStream();
    ImageOutputStream stream = createOutputStream(baStream);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setHeaderComment("Best crystal ever!");
    CbfImageWriter writer = createWriter(stream);
    writer.write(null, create1x1Image(mdata), null);
    String expected = "# Comment: Best crystal ever!\r\n";
    String actual = extractHeader(baStream.toByteArray());
    assertEquals(expected, actual);
  }

  /*
   * Source region and 2x2 subsampling with offset with byte_offset
   * compression.
   */
  @Test
  public void testWrite01() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 24, 26 },
      { 36, 38 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(1, 1, 4, 4));
    param.setSourceSubsampling(2, 2, 1, 1);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Source region and 3x1 subsampling with offset with byte_offset
   * compression.
   */
  @Test
  public void testWrite02() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 18 },
      { 24 },
      { 30 },
      { 36 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(1, 1, 4, 4));
    param.setSourceSubsampling(3, 1, 1, 0);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region and 2x2 subsampling with byte_offset compression. */
  @Test
  public void testWrite03() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 19, 21 },
      { 31, 33 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(3, 1, 3, 4));
    param.setSourceSubsampling(2, 2, 0, 0);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region and 3x2 subsampling with byte_offset compression. */
  @Test
  public void testWrite04() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 23, 26 },
      { 35, 38 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(1, 2, 4, 3));
    param.setSourceSubsampling(3, 2, 0, 0);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region with byte_offset compression. */
  @Test
  public void testWrite05() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 18, 19, 20 },
      { 24, 25, 26 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(2, 1, 3, 2));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region is entire image with byte_offset compression. */
  @Test
  public void testWrite06() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(0, 0, output[0].length, output.length));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Signed 32-bit integer with byte_offset compression. */
  @Test
  public void testWrite07() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage));
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Binary image stored as 8-bit unsigned integer with byte_offset
   * compression.
   */
  @Test
  public void testWrite08() throws IOException {
    byte[][] output = {
      { 0, 1, 1 },
      { 1, 0, 1 }
    };
    byte[][] expected = output;

    BufferedImage outImage = createByteBinaryBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_8_BIT_INTEGER);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_BYTE_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_BYTE, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getBytePixels(inImage));
  }

  /* Unsigned 8-bit integer with byte_offset compression. */
  @Test
  public void testWrite09() throws IOException {
    short[][] output = {
      { Byte.MAX_VALUE + 3, 6, 7 },
      { Byte.MAX_VALUE + 5, 8, 9 }
    };
    short[][] expected = output;

    BufferedImage outImage = createUbyteBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_8_BIT_INTEGER);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_BYTE_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_BYTE, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getShortPixels(inImage));
  }

  /* Signed 8-bit integer with byte_offset compression. */
  @Test
  public void testWrite10() throws IOException {
    byte[][] output = {
      { -1, -2, -3 },
      { +4, +5, +6 }
    };
    byte[][] expected = output;

    BufferedImage outImage = createByteBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.SIGNED_8_BIT_INTEGER);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_SHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getBytePixels(inImage));
  }

  /*
   * Signed 8-bit integer, without specified element type, with byte_offset
   * compression.
   */
  @Test
  public void testWrite11() throws IOException {
    byte[][] output = {
      { -1, -2, -3 },
      { +4, +5, +6 }
    };
    byte[][] expected = output;

    BufferedImage outImage = createByteBufferedImage(output);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_SHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getBytePixels(inImage));
  }

  /* Unsigned 16-bit integer with byte_offset compression. */
  @Test
  public void testWrite12() throws IOException {
    int[][] output = {
      { Short.MAX_VALUE + 3, 6, 7 },
      { Short.MAX_VALUE + 5, 8, 9 }
    };
    int[][] expected = output;

    BufferedImage outImage = createUshortBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_16_BIT_INTEGER);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_USHORT_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_USHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Signed 16-bit integer with byte_offset compression. */
  @Test
  public void testWrite13() throws IOException {
    short[][] output = {
      { Short.MAX_VALUE, +1, +2, +3 },
      { Short.MIN_VALUE, -1, -2, -3 }
    };
    short[][] expected = output;

    BufferedImage outImage = createShortBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.SIGNED_16_BIT_INTEGER);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_SHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getShortPixels(inImage));
  }

  /*
   * Signed 16-bit integer, without specified element type, with byte_offset
   * compression.
   */
  @Test
  public void testWrite14() throws IOException {
    short[][] output = {
      { Short.MAX_VALUE, +1, +2, +3 },
      { Short.MIN_VALUE, -1, -2, -3 }
    };
    short[][] expected = output;

    BufferedImage outImage = createShortBufferedImage(output);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_SHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getShortPixels(inImage));
  }

  /* Float as unsigned 32-bit integer with byte_offset compression. */
  @Test
  public void testWrite15() throws IOException {
    float[][] output = {
      { Integer.MAX_VALUE + 3.0f, 6.0f, Integer.MAX_VALUE + 50.0f },
      { Integer.MAX_VALUE + 5.0f, 8.0f, Integer.MAX_VALUE + 51.0f }
    };
    float[][] expected = output;

    BufferedImage outImage = createFloatBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.UNSIGNED_32_BIT_INTEGER);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.UNSIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_FLOAT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /* Float as signed 32-bit integer with byte_offset compression. */
  @Test
  public void testWrite16() throws IOException {
    float[][] output = {
      { Integer.MAX_VALUE, -6.0f, Integer.MAX_VALUE - 50.0f },
      { Integer.MIN_VALUE, +8.0f, Integer.MAX_VALUE - 51.0f }
    };
    float[][] expected = output;

    BufferedImage outImage = createFloatBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /* Int as unsigned 32-bit integer with byte_offset compression. */
  @Test
  public void testWrite17() throws IOException {
    int[][] output = {
      { Integer.MAX_VALUE, 6, 7 },
      { Integer.MAX_VALUE, 0, Integer.MAX_VALUE - 1 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.UNSIGNED_32_BIT_INTEGER);
    byte[] data = writeBufferedImage(outImage, outMdata);
    CbfImageReader reader = createReader();
    reader.setInput(createInputStream(data));
    ImageReadParam param = reader.getDefaultReadParam();
    boolean foundIntType = false;
    for (Iterator<ImageTypeSpecifier> i = reader.getImageTypes(0); i.hasNext();) {
      ImageTypeSpecifier each = i.next();
      if (each.getSampleModel().getDataType() != DataBuffer.TYPE_INT) continue;
      param.setDestinationType(each);
      foundIntType = true;
    }
    assertTrue(foundIntType);
    BufferedImage inImage = reader.read(0, param);
    CbfMetadata inMdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(CbfElementType.UNSIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Int as signed 32-bit integer with byte_offset compression. */
  @Test
  public void testWrite18() throws IOException {
    int[][] output = {
      { Integer.MAX_VALUE, 6, Integer.MAX_VALUE - 1 },
      { Integer.MIN_VALUE, 8, Integer.MIN_VALUE + 1 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Int as signed 32-bit integer, without specified element type, with
   * byte_offset compression.
   */
  @Test
  public void testWrite19() throws IOException {
    int[][] output = {
      { Integer.MAX_VALUE, 6, Integer.MAX_VALUE - 1 },
      { Integer.MIN_VALUE, 8, Integer.MIN_VALUE + 1 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    Object[] inImageAndMetadata = readBufferedImageAndMetadata(writeBufferedImage(outImage));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Float as signed 32-bit integer with byte_offset compression. */
  @Test
  public void testWrite20() throws IOException {
    float[][] output = {
      { Integer.MAX_VALUE, 6, 7 },
      { Integer.MAX_VALUE, 0, Integer.MAX_VALUE - 1 }
    };
    float[][] expected = output;

    BufferedImage outImage = createFloatBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    byte[] data = writeBufferedImage(outImage, outMdata);
    CbfImageReader reader = createReader();
    reader.setInput(createInputStream(data));
    ImageReadParam param = reader.getDefaultReadParam();
    boolean foundFloatType = false;
    for (Iterator<ImageTypeSpecifier> i = reader.getImageTypes(0); i.hasNext();) {
      ImageTypeSpecifier each = i.next();
      if (each.getSampleModel().getDataType() != DataBuffer.TYPE_FLOAT) continue;
      param.setDestinationType(each);
      foundFloatType = true;
    }
    assertTrue(foundFloatType);
    BufferedImage inImage = reader.read(0, param);
    CbfMetadata inMdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_FLOAT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /* Source region and 2x2 subsampling with offset with none compression. */
  @Test
  public void testWrite21() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 24, 26 },
      { 36, 38 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(1, 1, 4, 4));
    param.setSourceSubsampling(2, 2, 1, 1);
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region and 3x1 subsampling with offset with none compression. */
  @Test
  public void testWrite22() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 18 },
      { 24 },
      { 30 },
      { 36 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(1, 1, 4, 4));
    param.setSourceSubsampling(3, 1, 1, 0);
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region and 2x2 subsampling with none compression. */
  @Test
  public void testWrite23() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 19, 21 },
      { 31, 33 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(3, 1, 3, 4));
    param.setSourceSubsampling(2, 2, 0, 0);
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region and 3x2 subsampling with none compression. */
  @Test
  public void testWrite24() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 23, 26 },
      { 35, 38 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(1, 2, 4, 3));
    param.setSourceSubsampling(3, 2, 0, 0);
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region with none compression. */
  @Test
  public void testWrite25() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 18, 19, 20 },
      { 24, 25, 26 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(2, 1, 3, 2));
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region is entire image with none compression. */
  @Test
  public void testWrite26() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setSourceRegion(new Rectangle(0, 0, output[0].length, output.length));
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Signed 32-bit integer with none compression. */
  @Test
  public void testWrite27() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Binary image stored as 8-bit unsigned integer with none compression. */
  @Test
  public void testWrite28() throws IOException {
    byte[][] output = {
      { 0, 1, 1 },
      { 1, 0, 1 }
    };
    byte[][] expected = output;

    BufferedImage outImage = createByteBinaryBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_8_BIT_INTEGER);
    mdata.setDataCompression(CbfCompression.NONE);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_BYTE_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_BYTE, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getBytePixels(inImage));
  }

  /* Unsigned 8-bit integer with none compression. */
  @Test
  public void testWrite29() throws IOException {
    short[][] output = {
      { Byte.MAX_VALUE + 3, 6, 7 },
      { Byte.MAX_VALUE + 5, 8, 9 }
    };
    short[][] expected = output;

    BufferedImage outImage = createUbyteBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_8_BIT_INTEGER);
    mdata.setDataCompression(CbfCompression.NONE);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_BYTE_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_BYTE, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getShortPixels(inImage));
  }

  /* Signed 8-bit integer with none compression. */
  @Test
  public void testWrite30() throws IOException {
    byte[][] output = {
      { -1, -2, -3 },
      { +4, +5, +6 }
    };
    byte[][] expected = output;

    BufferedImage outImage = createByteBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.SIGNED_8_BIT_INTEGER);
    mdata.setDataCompression(CbfCompression.NONE);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_SHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getBytePixels(inImage));
  }

  /*
   * Signed 8-bit integer, without specified element type, with none
   * compression.
   */
  @Test
  public void testWrite31() throws IOException {
    byte[][] output = {
      { -1, -2, -3 },
      { +4, +5, +6 }
    };
    byte[][] expected = output;

    BufferedImage outImage = createByteBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_SHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getBytePixels(inImage));
  }

  /* Unsigned 16-bit integer with none compression. */
  @Test
  public void testWrite32() throws IOException {
    int[][] output = {
      { Short.MAX_VALUE + 3, 6, 7 },
      { Short.MAX_VALUE + 5, 8, 9 }
    };
    int[][] expected = output;

    BufferedImage outImage = createUshortBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_16_BIT_INTEGER);
    mdata.setDataCompression(CbfCompression.NONE);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_USHORT_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_USHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Signed 16-bit integer with none compression. */
  @Test
  public void testWrite33() throws IOException {
    short[][] output = {
      { Short.MAX_VALUE, +1, +2, +3 },
      { Short.MIN_VALUE, -1, -2, -3 }
    };
    short[][] expected = output;

    BufferedImage outImage = createShortBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.SIGNED_16_BIT_INTEGER);
    mdata.setDataCompression(CbfCompression.NONE);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_SHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getShortPixels(inImage));
  }

  /*
   * Signed 16-bit integer, without specified element type, with none
   * compression.
   */
  @Test
  public void testWrite34() throws IOException {
    short[][] output = {
      { Short.MAX_VALUE, +1, +2, +3 },
      { Short.MIN_VALUE, -1, -2, -3 }
    };
    short[][] expected = output;

    BufferedImage outImage = createShortBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, param));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_SHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getShortPixels(inImage));
  }

  /* Float as unsigned 32-bit integer with none compression. */
  @Test
  public void testWrite35() throws IOException {
    float[][] output = {
      { Integer.MAX_VALUE + 3.0f, 6.0f, Integer.MAX_VALUE + 50.0f },
      { Integer.MAX_VALUE + 5.0f, 8.0f, Integer.MAX_VALUE + 51.0f }
    };
    float[][] expected = output;

    BufferedImage outImage = createFloatBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.UNSIGNED_32_BIT_INTEGER);
    outMdata.setDataCompression(CbfCompression.NONE);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.UNSIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_FLOAT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /* Float as signed 32-bit integer with none compression. */
  @Test
  public void testWrite36() throws IOException {
    float[][] output = {
      { Integer.MAX_VALUE, -6.0f, Integer.MAX_VALUE - 50.0f },
      { Integer.MIN_VALUE, +8.0f, Integer.MAX_VALUE - 51.0f }
    };
    float[][] expected = output;

    BufferedImage outImage = createFloatBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    outMdata.setDataCompression(CbfCompression.NONE);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /* Int as unsigned 32-bit integer with none compression. */
  @Test
  public void testWrite37() throws IOException {
    int[][] output = {
      { Integer.MAX_VALUE, 6, 7 },
      { Integer.MAX_VALUE, 0, Integer.MAX_VALUE - 1 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.UNSIGNED_32_BIT_INTEGER);
    outMdata.setDataCompression(CbfCompression.NONE);
    byte[] data = writeBufferedImage(outImage, outMdata);
    CbfImageReader reader = createReader();
    reader.setInput(createInputStream(data));
    ImageReadParam param = reader.getDefaultReadParam();
    boolean foundIntType = false;
    for (Iterator<ImageTypeSpecifier> i = reader.getImageTypes(0); i.hasNext();) {
      ImageTypeSpecifier each = i.next();
      if (each.getSampleModel().getDataType() != DataBuffer.TYPE_INT) continue;
      param.setDestinationType(each);
      foundIntType = true;
    }
    assertTrue(foundIntType);
    BufferedImage inImage = reader.read(0, param);
    CbfMetadata inMdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(CbfElementType.UNSIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Int as signed 32-bit integer with none compression. */
  @Test
  public void testWrite38() throws IOException {
    int[][] output = {
      { Integer.MAX_VALUE, 6, Integer.MAX_VALUE - 1 },
      { Integer.MIN_VALUE, 8, Integer.MIN_VALUE + 1 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    outMdata.setDataCompression(CbfCompression.NONE);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Int as signed 32-bit integer, without specified element type, with none
   * compression.
   */
  @Test
  public void testWrite39() throws IOException {
    int[][] output = {
      { Integer.MAX_VALUE, 6, Integer.MAX_VALUE - 1 },
      { Integer.MIN_VALUE, 8, Integer.MIN_VALUE + 1 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, param));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Float as signed 32-bit integer with none compression. */
  @Test
  public void testWrite40() throws IOException {
    float[][] output = {
      { Integer.MAX_VALUE, 6, 7 },
      { Integer.MAX_VALUE, 0, Integer.MAX_VALUE - 1 }
    };
    float[][] expected = output;

    BufferedImage outImage = createFloatBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    outMdata.setDataCompression(CbfCompression.NONE);
    byte[] data = writeBufferedImage(outImage, outMdata);
    CbfImageReader reader = createReader();
    reader.setInput(createInputStream(data));
    ImageReadParam param = reader.getDefaultReadParam();
    boolean foundFloatType = false;
    for (Iterator<ImageTypeSpecifier> i = reader.getImageTypes(0); i.hasNext();) {
      ImageTypeSpecifier each = i.next();
      if (each.getSampleModel().getDataType() != DataBuffer.TYPE_FLOAT) continue;
      param.setDestinationType(each);
      foundFloatType = true;
    }
    assertTrue(foundFloatType);
    BufferedImage inImage = reader.read(0, param);
    CbfMetadata inMdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_FLOAT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /* Float as signed 32-bit real IEEE with none compression. */
  @Test
  public void testWrite41() throws IOException {
    float[][] output = {
      { Float.MAX_VALUE, 6, Float.MAX_VALUE - 1 },
      { Float.MIN_VALUE, 8, Float.MIN_VALUE + 1 },
      { 1.234331235e20f, 8, -1.23423123925e-20f }
    };
    float[][] expected = output;

    BufferedImage outImage = createFloatBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_REAL_IEEE);
    outMdata.setDataCompression(CbfCompression.NONE);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_REAL_IEEE, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_FLOAT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /*
   * Float as signed 32-bit real IEEE, without specified element type, with
   * none compression.
   */
  @Test
  public void testWrite42() throws IOException {
    float[][] output = {
      { Float.MAX_VALUE, 6, Float.MAX_VALUE - 1 },
      { Float.MIN_VALUE, 8, Float.MIN_VALUE + 1 },
      { 1.234331235e20f, 8, -1.23423123925e-20f }
    };
    float[][] expected = output;

    BufferedImage outImage = createFloatBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, param));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_REAL_IEEE, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_FLOAT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /* Double as signed 64-bit real IEEE with none compression. */
  @Test
  public void testWrite43() throws IOException {
    double[][] output = {
      { Double.MAX_VALUE, 6, Double.MAX_VALUE - 1 },
      { Double.MIN_VALUE, 8, Double.MIN_VALUE + 1 },
      { 1.2343423235e201, 8, -1.234342129251e-201 }
    };
    double[][] expected = output;

    BufferedImage outImage = createDoubleBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_64_BIT_REAL_IEEE);
    outMdata.setDataCompression(CbfCompression.NONE);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_64_BIT_REAL_IEEE, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_DOUBLE, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getDoublePixels(inImage));
  }

  /*
   * Double as signed 64-bit real IEEE, without specified element type, with
   * none compression.
   */
  @Test
  public void testWrite44() throws IOException {
    double[][] output = {
      { Double.MAX_VALUE, 6, Double.MAX_VALUE - 1 },
      { Double.MIN_VALUE, 8, Double.MIN_VALUE + 1 },
      { 1.2343423235e201, 8, -1.234342129251e-201 }
    };
    double[][] expected = output;

    BufferedImage outImage = createDoubleBufferedImage(output);
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, param));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_64_BIT_REAL_IEEE, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_DOUBLE, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getDoublePixels(inImage));
  }

  /* ImageWriteParam compression overrides CbfMetadata compression. */
  @Test
  public void testWrite45() throws IOException {
    byte[][] output = {
      { -1, -2, -3 },
      { +4, +5, +6 }
    };

    BufferedImage outImage = createByteBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataCompression(CbfCompression.NONE);
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
    param.setCompressionType(CbfImageWriteParam.COMPRESSION_BYTE_OFFSET);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, mdata, param));
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
  }

  /* ImageWriteParam compression overrides CbfMetadata compression. */
  @Test
  public void testWrite46() throws IOException {
    byte[][] output = {
      { -1, -2, -3 },
      { +4, +5, +6 }
    };

    BufferedImage outImage = createByteBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataCompression(CbfCompression.BYTE_OFFSET);
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
    param.setCompressionType(CbfImageWriteParam.COMPRESSION_NONE);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, mdata, param));
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
  }

  /* Progress for byte_offset compression. */
  @Test
  public void testWrite47() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    final List<Boolean> started = Arrays.asList(new Boolean[] { Boolean.FALSE });
    final List<Float> percentDone = new ArrayList<Float>();
    final List<Boolean> completed = Arrays.asList(new Boolean[] { Boolean.FALSE });
    final List<Boolean> aborted = Arrays.asList(new Boolean[] { Boolean.FALSE });
    IIOWriteProgressListener listener = new IIOWriteProgressListener() {
      public void imageStarted(ImageWriter source, int imageIndex) {
        started.set(0, Boolean.TRUE);
      }

      public void imageProgress(ImageWriter source, float percentageDone) {
        percentDone.add(percentageDone);
      }

      public void imageComplete(ImageWriter source) {
        completed.set(0, Boolean.TRUE);
      }

      public void thumbnailStarted(ImageWriter source, int imageIndex, int thumbnailIndex) {
      }

      public void thumbnailProgress(ImageWriter source, float percentageDone) {
      }

      public void thumbnailComplete(ImageWriter source) {
      }

      public void writeAborted(ImageWriter source) {
        aborted.set(0, Boolean.TRUE);
      }
    };
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
    param.setCompressionType(CbfImageWriteParam.COMPRESSION_BYTE_OFFSET);
    writeBufferedImage(outImage, param, listener);
    assertTrue(started.get(0));
    assertTrue(completed.get(0));
    assertFalse(aborted.get(0));
    assertEquals(5, percentDone.size());
    assertEquals(80.0f, percentDone.get(percentDone.size() - 1), 0.1f);
  }

  /* Progress for none compression. */
  @Test
  public void testWrite48() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    final List<Boolean> started = Arrays.asList(new Boolean[] { Boolean.FALSE });
    final List<Float> percentDone = new ArrayList<Float>();
    final List<Boolean> completed = Arrays.asList(new Boolean[] { Boolean.FALSE });
    final List<Boolean> aborted = Arrays.asList(new Boolean[] { Boolean.FALSE });
    IIOWriteProgressListener listener = new IIOWriteProgressListener() {
      public void imageStarted(ImageWriter source, int imageIndex) {
        started.set(0, Boolean.TRUE);
      }

      public void imageProgress(ImageWriter source, float percentageDone) {
        percentDone.add(percentageDone);
      }

      public void imageComplete(ImageWriter source) {
        completed.set(0, Boolean.TRUE);
      }

      public void thumbnailStarted(ImageWriter source, int imageIndex, int thumbnailIndex) {
      }

      public void thumbnailProgress(ImageWriter source, float percentageDone) {
      }

      public void thumbnailComplete(ImageWriter source) {
      }

      public void writeAborted(ImageWriter source) {
        aborted.set(0, Boolean.TRUE);
      }
    };
    ImageWriteParam param = createImageWriteParam();
    param.setCompressionMode(ImageWriteParam.MODE_DISABLED);
    writeBufferedImage(outImage, param, listener);
    assertTrue(started.get(0));
    assertTrue(completed.get(0));
    assertFalse(aborted.get(0));
    assertEquals(5, percentDone.size());
    assertEquals(80.0f, percentDone.get(percentDone.size() - 1), 0.1f);
  }

  /* Int as signed 32-bit integer with byte_offset compression. */
  @Test
  public void testWrite49() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, Integer.MAX_VALUE },
      { 16, 17, 18, 19, 20, Integer.MIN_VALUE },
      { 22, 23, 24, 25, 26, Short.MAX_VALUE + 51 },
      { 28, 29, 30, 31, 32, 33 },
      { 28, 29, 30, 31, 32, 34 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Second band (index 1) of RGB source image as signed 32-bit integer with
   * byte_offset compression.
   */
  @Test
  public void testWrite50() throws IOException {
    int[][][] output = {
      { { 0, 0, 0 }, { 0, +1, 0 }, { 0, +2, 0 } },
      { { 0, 3, 0 }, { 0, +4, 0 }, { 0, +5, 0 } },
      { { 0, 6, 0 }, { 0, +7, 0 }, { 0, +8, 0 } },
      { { 0, 9, 0 }, { 0, 10, 0 }, { 0, 11, 0 } }
    };
    int[][] expected = new int[output.length][output[0].length];
    for (int r = 0; r < expected.length; r++) {
      for (int c = 0; c < expected[0].length; c++) {
        expected[r][c] = output[r][c][1];
      }
    }

    BufferedImage outImage = createIntRgbBufferedImage(output);
    ImageWriteParam outParam = createImageWriteParam();
    outParam.setSourceBands(new int[] { 1 });
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, outMdata, outParam));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Second band (index 1) of RGB source image as signed 32-bit integer with
   * none compression.
   */
  @Test
  public void testWrite51() throws IOException {
    int[][][] output = {
      { { 0, 0, 0 }, { 0, +1, 0 }, { 0, +2, 0 } },
      { { 0, 3, 0 }, { 0, +4, 0 }, { 0, +5, 0 } },
      { { 0, 6, 0 }, { 0, +7, 0 }, { 0, +8, 0 } },
      { { 0, 9, 0 }, { 0, 10, 0 }, { 0, 11, 0 } }
    };
    int[][] expected = new int[output.length][output[0].length];
    for (int r = 0; r < expected.length; r++) {
      for (int c = 0; c < expected[0].length; c++) {
        expected[r][c] = output[r][c][1];
      }
    }

    BufferedImage outImage = createIntRgbBufferedImage(output);
    ImageWriteParam outParam = createImageWriteParam();
    outParam.setSourceBands(new int[] { 1 });
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    outMdata.setDataCompression(CbfCompression.NONE);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, outMdata, outParam));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Float, including negative values, as unsigned 32-bit integer with
   * byte_offset compression. Negative values are clipped to 0.
   */
  @Test
  public void testWrite52() throws IOException {
    float[][] output = {
      { Integer.MIN_VALUE - 40.0f, -6.0f, Integer.MAX_VALUE + 50.0f },
      { Integer.MIN_VALUE - 41.0f, +8.0f, Integer.MAX_VALUE + 51.0f }
    };
    float[][] expected = {
      { 0.0f, 0.0f, Integer.MAX_VALUE + 50.0f },
      { 0.0f, 8.0f, Integer.MAX_VALUE + 51.0f }
    };

    BufferedImage outImage = createFloatBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.UNSIGNED_32_BIT_INTEGER);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, outMdata));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_FLOAT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /*
   * Int, including negative values, as unsigned 32-bit integer with
   * byte_offset compression. Negative values are clipped to 0.
   */
  @Test
  public void testWrite53() throws IOException {
    int[][] output = {
      { Integer.MIN_VALUE + 0, -6, -1 },
      { Integer.MIN_VALUE + 1, +0, +7 }
    };
    int[][] expected = {
      { 0, 0, 0 },
      { 0, 0, 7 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.UNSIGNED_32_BIT_INTEGER);
    byte[] data = writeBufferedImage(outImage, outMdata);
    CbfImageReader reader = createReader();
    reader.setInput(createInputStream(data));
    ImageReadParam param = reader.getDefaultReadParam();
    boolean foundIntType = false;
    for (Iterator<ImageTypeSpecifier> i = reader.getImageTypes(0); i.hasNext();) {
      ImageTypeSpecifier each = i.next();
      if (each.getSampleModel().getDataType() != DataBuffer.TYPE_INT) continue;
      param.setDestinationType(each);
      foundIntType = true;
    }
    assertTrue(foundIntType);
    BufferedImage inImage = reader.read(0, param);
    CbfMetadata inMdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(CbfElementType.UNSIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Float, including negative values, as unsigned 32-bit integer with none
   * compression. Negative values are clipped to 0.
   */
  @Test
  public void testWrite54() throws IOException {
    float[][] output = {
      { Integer.MIN_VALUE - 40.0f, -6.0f, Integer.MAX_VALUE + 50.0f },
      { Integer.MIN_VALUE - 41.0f, +8.0f, Integer.MAX_VALUE + 51.0f }
    };
    float[][] expected = {
      { 0.0f, 0.0f, Integer.MAX_VALUE + 50.0f },
      { 0.0f, 8.0f, Integer.MAX_VALUE + 51.0f }
    };

    BufferedImage outImage = createFloatBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.UNSIGNED_32_BIT_INTEGER);
    outMdata.setDataCompression(CbfCompression.NONE);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, outMdata));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_FLOAT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getFloatPixels(inImage));
  }

  /*
   * Int, including negative values, as unsigned 32-bit integer with none
   * compression. Negative values are clipped to 0.
   */
  @Test
  public void testWrite55() throws IOException {
    int[][] output = {
      { Integer.MIN_VALUE + 0, -6, -1 },
      { Integer.MIN_VALUE + 1, +0, +7 }
    };
    int[][] expected = {
      { 0, 0, 0 },
      { 0, 0, 7 }
    };

    BufferedImage outImage = createIntBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.UNSIGNED_32_BIT_INTEGER);
    outMdata.setDataCompression(CbfCompression.NONE);
    byte[] data = writeBufferedImage(outImage, outMdata);
    CbfImageReader reader = createReader();
    reader.setInput(createInputStream(data));
    ImageReadParam param = reader.getDefaultReadParam();
    boolean foundIntType = false;
    for (Iterator<ImageTypeSpecifier> i = reader.getImageTypes(0); i.hasNext();) {
      ImageTypeSpecifier each = i.next();
      if (each.getSampleModel().getDataType() != DataBuffer.TYPE_INT) continue;
      param.setDestinationType(each);
      foundIntType = true;
    }
    assertTrue(foundIntType);
    BufferedImage inImage = reader.read(0, param);
    CbfMetadata inMdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(CbfElementType.UNSIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.NONE, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Short, including negative values, as unsigned 16-bit integer with
   * byte_offset compression. Negative values are clipped to 0.
   */
  @Test
  public void testWrite56() throws IOException {
    short[][] output = {
      { Short.MIN_VALUE + 0, 0, +1, +2, +3 },
      { Short.MIN_VALUE + 1, 0, -1, -2, -3 }
    };
    int[][] expected = {
      { 0, 0, +1, +2, +3 },
      { 0, 0, +0, +0, +0 }
    };

    BufferedImage outImage = createShortBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_16_BIT_INTEGER);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_USHORT_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_USHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Short, including negative values, as unsigned 16-bit integer with none
   * compression. Negative values are clipped to 0.
   */
  @Test
  public void testWrite57() throws IOException {
    short[][] output = {
      { Short.MIN_VALUE + 0, 0, +1, +2, +3 },
      { Short.MIN_VALUE + 1, 0, -1, -2, -3 }
    };
    int[][] expected = {
      { 0, 0, +1, +2, +3 },
      { 0, 0, +0, +0, +0 }
    };

    BufferedImage outImage = createShortBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_16_BIT_INTEGER);
    mdata.setDataCompression(CbfCompression.NONE);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_USHORT_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_USHORT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Byte, including negative values, as unsigned 8-bit integer with
   * byte_offset compression. Negative values are clipped to 0.
   */
  @Test
  public void testWrite58() throws IOException {
    byte[][] output = {
      { Byte.MIN_VALUE + 0, 0, +1, +2, +3 },
      { Byte.MIN_VALUE + 1, 0, -1, -2, -3 }
    };
    short[][] expected = {
      { 0, 0, +1, +2, +3 },
      { 0, 0, +0, +0, +0 }
    };

    BufferedImage outImage = createByteBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_8_BIT_INTEGER);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_BYTE_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_BYTE, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getShortPixels(inImage));
  }

  /*
   * Byte, including negative values, as unsigned 16-bit integer with none
   * compression. Negative values are clipped to 0.
   */
  @Test
  public void testWrite59() throws IOException {
    byte[][] output = {
      { Byte.MIN_VALUE + 0, 0, +1, +2, +3 },
      { Byte.MIN_VALUE + 1, 0, -1, -2, -3 }
    };
    short[][] expected = {
      { 0, 0, +1, +2, +3 },
      { 0, 0, +0, +0, +0 }
    };

    BufferedImage outImage = createByteBufferedImage(output);
    CbfMetadata mdata = new CbfMetadata();
    mdata.setDataElementType(CbfElementType.UNSIGNED_8_BIT_INTEGER);
    mdata.setDataCompression(CbfCompression.NONE);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage, mdata));
    assertEquals(BufferedImage.TYPE_BYTE_GRAY, inImage.getType());
    assertEquals(DataBuffer.TYPE_BYTE, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getShortPixels(inImage));
  }

  /* Null data element type should default to signed 32-bit integer. */
  @Test
  public void testWrite60() throws IOException {
    int[][] output = {
      { Integer.MAX_VALUE, 6, Integer.MAX_VALUE - 1 },
      { Integer.MIN_VALUE, 8, Integer.MIN_VALUE + 1 }
    };
    int[][] expected = output;

    BufferedImage outImage = createIntBufferedImage(output);
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(null);
    Object[] inImageAndMetadata =
      readBufferedImageAndMetadata(writeBufferedImage(outImage, outMdata));
    BufferedImage inImage = (BufferedImage)inImageAndMetadata[0];
    CbfMetadata inMdata = (CbfMetadata)inImageAndMetadata[1];
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, inMdata.getDataElementType());
    assertEquals(CbfCompression.BYTE_OFFSET, inMdata.getDataCompression());
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  private ImageOutputStream createOutputStream(OutputStream stream) {
    return new MemoryCacheImageOutputStream(stream);
  }

  private CbfImageWriter createWriter(ImageOutputStream stream) throws IOException {
    CbfImageWriterSpi spi = new CbfImageWriterSpi();
    CbfImageWriter writer = (CbfImageWriter)spi.createWriterInstance();
    writer.setOutput(stream);
    return writer;
  }

  private IIOImage create1x1Image(IIOMetadata metadata) {
    return new IIOImage(new BufferedImage(1, 1, BufferedImage.TYPE_BYTE_BINARY), null, metadata);
  }

  private String extractHeader(byte[] image) {
    try {
      InputStream stream = new ByteArrayInputStream(image);
      BufferedReader reader = new BufferedReader(new InputStreamReader(stream, "US-ASCII"));
      StringBuffer result = new StringBuffer();
      String line = reader.readLine();
      while (line != null && !line.equals("_array_data.header_contents")) {
        line = reader.readLine();
      }
      line = reader.readLine();
      if (line == null || !line.equals(";")) fail("Did not find ';' line at start of header");
      line = reader.readLine();
      while (line != null && !line.equals(";")) {
        result.append(line + "\r\n");
        line = reader.readLine();
      }
      return result.toString();
    } catch (Exception e) {
      fail(e.toString());
      return null;
    }
  }

  private CbfMetadata createMetadata() {
    CbfMetadata result = new CbfMetadata();
    result.setHeaderDefaults();
    Calendar c = Calendar.getInstance(Locale.US);
    c.set(Calendar.YEAR, 2010);
    c.set(Calendar.MONTH, 10);
    c.set(Calendar.DAY_OF_MONTH, 17);
    c.set(Calendar.HOUR_OF_DAY, 15);
    c.set(Calendar.MINUTE, 52);
    c.set(Calendar.SECOND, 30);
    c.set(Calendar.MILLISECOND, 834);
    result.setHeaderDate(c.getTime());
    return result;
  }
}
