/*
 * Copyright (c) 2011 J. Lewis Muir <jlmuir@imca.aps.anl.gov>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package ch.psi.imagej.cbf;

import static org.junit.Assert.fail;

import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import javax.imageio.IIOImage;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.event.IIOReadProgressListener;
import javax.imageio.event.IIOWriteProgressListener;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.ImageOutputStream;
import javax.imageio.stream.MemoryCacheImageInputStream;
import javax.imageio.stream.MemoryCacheImageOutputStream;

import ch.psi.imagej.cbf.CbfImageReader;
import ch.psi.imagej.cbf.CbfImageReaderSpi;
import ch.psi.imagej.cbf.CbfImageWriter;
import ch.psi.imagej.cbf.CbfImageWriterSpi;
import ch.psi.imagej.cbf.CbfMetadata;

/**
 * I provide utility methods for use in testing.
 */
public class TestUtils {
  public static ImageInputStream streamFromString(String s) {
    return streamFromBytes(asciiBytes(s));
  }

  public static ImageInputStream streamFromBytes(byte[] b) {
    return new MemoryCacheImageInputStream(new ByteArrayInputStream(b));
  }

  public static byte[] asciiBytes(String value) {
    try {
      return value.getBytes("US-ASCII");
    } catch (UnsupportedEncodingException e) {
      fail(e.toString());
      return new byte[] {};
    }
  }

  protected static String createImageText() {
    return "###CBF: VERSION 1.5, CBFlib v0.7.8 - SLS/DECTRIS PILATUS detectors\r\n"
      + "\r\n"
      + "data_thaumatin_0001\r\n"
      + "\r\n"
      + "_array_data.header_convention \"SLS_1.0\"\r\n"
      + "_array_data.header_contents\r\n"
      + ";\r\n"
      + "# Detector: PILATUS 6M, 60-0103, IMCA-CAT\r\n"
      + "# 2010-Nov-17T15:52:30.834\r\n"
      + "# Pixel_size 172e-6 m x 172e-6 m\r\n"
      + "# Silicon sensor, thickness 0.000320 m\r\n"
      + "# Exposure_time 0.081033 s\r\n"
      + "# Exposure_period 0.083333 s\r\n"
      + "# Tau = 124.0e-09 s\r\n"
      + "# Count_cutoff 326741 counts\r\n"
      + "# Threshold_setting 7439 eV\r\n"
      + "# Gain_setting not implemented (vrf = 9.900)\r\n"
      + "# N_excluded_pixels = 2391\r\n"
      + "# Excluded_pixels: badpix_mask.tif\r\n"
      + "# Flat_field: (nil)\r\n"
      + "# Trim_directory: \r\n"
      + "# Wavelength 1.0000 A\r\n"
      + "# Energy_range (0, 0) eV\r\n"
      + "# Detector_distance 0.94997 m\r\n"
      + "# Detector_Voffset 0.00000 m\r\n"
      + "# Beam_xy (1253.00, 1288.00) pixels\r\n"
      + "# Flux 0.0000 ph/s\r\n"
      + "# Filter_transmission 0.0000\r\n"
      + "# Start_angle 0.0000 deg.\r\n"
      + "# Angle_increment 1.0000 deg.\r\n"
      + "# Detector_2theta 0.0007 deg.\r\n"
      + "# Polarization 0.990\r\n"
      + "# Alpha 0.0000 deg.\r\n"
      + "# Kappa 0.0000 deg.\r\n"
      + "# Phi 0.0000 deg.\r\n"
      + "# Chi 0.0000 deg.\r\n"
      + "# Oscillation_axis X, CW\r\n"
      + "# N_oscillations 1\r\n"
      + ";\r\n"
      + "\r\n"
      + "_array_data.data\r\n"
      + ";\r\n"
      + "--CIF-BINARY-FORMAT-SECTION--\r\n"
      + "Content-Type: application/octet-stream;\r\n"
      + "     conversions=\"x-CBF_BYTE_OFFSET\"\r\n"
      + "Content-Transfer-Encoding: BINARY\r\n"
      + "X-Binary-Size: 6224001\r\n"
      + "X-Binary-ID: 1\r\n"
      + "X-Binary-Element-Type: \"signed 32-bit integer\"\r\n"
      + "X-Binary-Element-Byte-Order: LITTLE_ENDIAN\r\n"
      + "Content-MD5: Ti67wd6I6DPuGIsOnMalMw==\r\n"
      + "X-Binary-Number-of-Elements: 6224001\r\n"
      + "X-Binary-Size-Fastest-Dimension: 2463\r\n"
      + "X-Binary-Size-Second-Dimension: 2527\r\n"
      + "X-Binary-Size-Padding: 4095\r\n"
      + "\r\n";
  }

  protected static byte[] createImageWithoutData() {
    return createImageWithoutData(asciiBytes(createImageText()));
  }

  protected static byte[] createImageWithoutData(byte[] text) {
    byte[] result = new byte[text.length + 4];
    System.arraycopy(text, 0, result, 0, text.length);
    result[result.length - 4] = 0x0C;
    result[result.length - 3] = 0x1A;
    result[result.length - 2] = 0x04;
    result[result.length - 1] = (byte)0xD5;
    return result;
  }

  protected static ImageInputStream createInputStream(byte[] cbf) {
    return new MemoryCacheImageInputStream(new ByteArrayInputStream(cbf));
  }

  protected static BufferedImage createByteBinaryBufferedImage(byte[][] pixels) {
    BufferedImage image = createByteBinaryBufferedImage(pixels[0].length, pixels.length);
    WritableRaster raster = image.getRaster();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        raster.setSample(x, y, 0, pixels[y][x]);
      }
    }
    return image;
  }

  protected static BufferedImage createByteBinaryBufferedImage(int width, int height) {
    return createByteBinarySpecifier().createBufferedImage(width, height);
  }

  protected static ImageTypeSpecifier createByteBinarySpecifier() {
    return ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_BYTE_BINARY);
  }

  protected static BufferedImage createUbyteBufferedImage(short[][] pixels) {
    BufferedImage image = createUbyteBufferedImage(pixels[0].length, pixels.length);
    WritableRaster raster = image.getRaster();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        raster.setSample(x, y, 0, pixels[y][x]);
      }
    }
    return image;
  }

  protected static BufferedImage createUbyteBufferedImage(int width, int height) {
    return createUbyteSpecifier().createBufferedImage(width, height);
  }

  protected static ImageTypeSpecifier createUbyteSpecifier() {
    return ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_BYTE_GRAY);
  }

  protected static BufferedImage createByteBufferedImage(byte[][] pixels) {
    BufferedImage image = createByteBufferedImage(pixels[0].length, pixels.length);
    WritableRaster raster = image.getRaster();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        raster.setSample(x, y, 0, pixels[y][x]);
      }
    }
    return image;
  }

  protected static BufferedImage createByteBufferedImage(int width, int height) {
    return createByteSpecifier().createBufferedImage(width, height);
  }

  protected static ImageTypeSpecifier createByteSpecifier() {
    /*
     * The type DataBuffer.TYPE_BYTE is for unsigned byte data so must use
     * DataBuffer.TYPE_SHORT. ImageTypeSpecifier.createGrayscale does not
     * support 8 bits with DataBuffer.TYPE_SHORT, so must use 16.
     */
    return ImageTypeSpecifier.createGrayscale(16, DataBuffer.TYPE_SHORT, true);
  }

  protected static BufferedImage createUshortBufferedImage(int[][] pixels) {
    BufferedImage image = createUshortBufferedImage(pixels[0].length, pixels.length);
    WritableRaster raster = image.getRaster();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        raster.setSample(x, y, 0, pixels[y][x]);
      }
    }
    return image;
  }

  protected static BufferedImage createUshortBufferedImage(int width, int height) {
    return createUshortSpecifier().createBufferedImage(width, height);
  }

  protected static ImageTypeSpecifier createUshortSpecifier() {
    return ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_USHORT_GRAY);
  }

  protected static BufferedImage createShortBufferedImage(short[][] pixels) {
    BufferedImage image = createShortBufferedImage(pixels[0].length, pixels.length);
    WritableRaster raster = image.getRaster();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        raster.setSample(x, y, 0, pixels[y][x]);
      }
    }
    return image;
  }

  protected static BufferedImage createShortBufferedImage(int width, int height) {
    return createShortSpecifier().createBufferedImage(width, height);
  }

  protected static ImageTypeSpecifier createShortSpecifier() {
    return ImageTypeSpecifier.createGrayscale(16, DataBuffer.TYPE_SHORT, true);
  }

  protected static BufferedImage createIntBufferedImage(int[][] pixels) {
    BufferedImage image = createIntBufferedImage(pixels[0].length, pixels.length);
    WritableRaster raster = image.getRaster();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        raster.setSample(x, y, 0, pixels[y][x]);
      }
    }
    return image;
  }

  protected static BufferedImage createIntBufferedImage(int width, int height) {
    return createIntSpecifier().createBufferedImage(width, height);
  }

  protected static ImageTypeSpecifier createIntSpecifier() {
    return ImageTypeSpecifier.createInterleaved(ColorSpace.getInstance(ColorSpace.CS_GRAY),
      new int[] { 0 }, DataBuffer.TYPE_INT, false, false);
  }

  protected static BufferedImage createIntRgbBufferedImage(int[][][] rgbPixels) {
    BufferedImage image = createIntRgbBufferedImage(rgbPixels[0].length, rgbPixels.length);
    WritableRaster raster = image.getRaster();
    for (int y = 0; y < rgbPixels.length; y++) {
      for (int x = 0; x < rgbPixels[0].length; x++) {
        raster.setPixel(x, y, rgbPixels[y][x]);
      }
    }
    return image;
  }

  protected static BufferedImage createIntRgbBufferedImage(int width, int height) {
    return createIntRgbSpecifier().createBufferedImage(width, height);
  }

  protected static ImageTypeSpecifier createIntRgbSpecifier() {
    return ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_INT_RGB);
  }

  protected static BufferedImage createFloatBufferedImage(float[][] pixels) {
    BufferedImage image = createFloatBufferedImage(pixels[0].length, pixels.length);
    WritableRaster raster = image.getRaster();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        raster.setSample(x, y, 0, pixels[y][x]);
      }
    }
    return image;
  }

  protected static BufferedImage createFloatBufferedImage(int width, int height) {
    return createFloatSpecifier().createBufferedImage(width, height);
  }

  protected static ImageTypeSpecifier createFloatSpecifier() {
    return ImageTypeSpecifier.createInterleaved(ColorSpace.getInstance(ColorSpace.CS_GRAY),
      new int[] { 0 }, DataBuffer.TYPE_FLOAT, false, false);
  }

  protected static BufferedImage createDoubleBufferedImage(double[][] pixels) {
    BufferedImage image = createDoubleBufferedImage(pixels[0].length, pixels.length);
    WritableRaster raster = image.getRaster();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        raster.setSample(x, y, 0, pixels[y][x]);
      }
    }
    return image;
  }

  protected static BufferedImage createDoubleBufferedImage(int width, int height) {
    return createDoubleSpecifier().createBufferedImage(width, height);
  }

  protected static ImageTypeSpecifier createDoubleSpecifier() {
    return ImageTypeSpecifier.createInterleaved(ColorSpace.getInstance(ColorSpace.CS_GRAY),
      new int[] { 0 }, DataBuffer.TYPE_DOUBLE, false, false);
  }

  protected static byte[][] getBytePixels(BufferedImage image) {
    byte[][] pixels = new byte[image.getHeight()][image.getWidth()];
    Raster raster = image.getData();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        pixels[y][x] = (byte)raster.getSample(x, y, 0);
      }
    }
    return pixels;
  }

  protected static short[][] getShortPixels(BufferedImage image) {
    short[][] pixels = new short[image.getHeight()][image.getWidth()];
    Raster raster = image.getData();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        pixels[y][x] = (short)raster.getSample(x, y, 0);
      }
    }
    return pixels;
  }

  protected static int[][] getIntPixels(BufferedImage image) {
    int[][] pixels = new int[image.getHeight()][image.getWidth()];
    Raster raster = image.getData();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        pixels[y][x] = raster.getSample(x, y, 0);
      }
    }
    return pixels;
  }

  protected static int[][][] getIntRgbPixels(BufferedImage image) {
    int[][][] pixels = new int[image.getHeight()][image.getWidth()][3];
    Raster raster = image.getData();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        raster.getPixel(x, y, pixels[y][x]);
      }
    }
    return pixels;
  }

  protected static float[][] getFloatPixels(BufferedImage image) {
    float[][] pixels = new float[image.getHeight()][image.getWidth()];
    Raster raster = image.getData();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        pixels[y][x] = raster.getSampleFloat(x, y, 0);
      }
    }
    return pixels;
  }

  protected static double[][] getDoublePixels(BufferedImage image) {
    double[][] pixels = new double[image.getHeight()][image.getWidth()];
    Raster raster = image.getData();
    for (int y = 0; y < pixels.length; y++) {
      for (int x = 0; x < pixels[0].length; x++) {
        pixels[y][x] = raster.getSampleDouble(x, y, 0);
      }
    }
    return pixels;
  }

  protected static BufferedImage readBufferedImage(byte[] input) throws IOException {
    return readBufferedImage(input, null, null);
  }

  protected static BufferedImage readBufferedImage(byte[] input, ImageReadParam param)
      throws IOException {
    return readBufferedImage(input, param, null);
  }

  protected static BufferedImage readBufferedImage(byte[] input, IIOReadProgressListener listener)
      throws IOException {
    return readBufferedImage(input, null, listener);
  }

  protected static BufferedImage readBufferedImage(byte[] input, ImageReadParam param,
      IIOReadProgressListener listener)
      throws IOException {
    return (BufferedImage)readBufferedImageAndMetadata(input, param, listener)[0];
  }

  protected static Object[] readBufferedImageAndMetadata(byte[] input) throws IOException {
    return readBufferedImageAndMetadata(input, null, null);
  }

  protected static Object[] readBufferedImageAndMetadata(byte[] input, ImageReadParam param,
      IIOReadProgressListener listener) throws IOException {
    ByteArrayInputStream inBytesStream = new ByteArrayInputStream(input);
    ImageInputStream inStream = new MemoryCacheImageInputStream(inBytesStream);
    CbfImageReader reader = createReader();
    if (listener != null) reader.addIIOReadProgressListener(listener);
    reader.setInput(inStream);
    BufferedImage result = reader.read(0, param);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    if (listener != null) reader.removeIIOReadProgressListener(listener);
    return new Object[] { result, mdata };
  }

  protected static CbfImageReader createReader() throws IOException {
    return (CbfImageReader)((new CbfImageReaderSpi()).createReaderInstance());
  }

  protected static byte[] writeBufferedImage(BufferedImage image) throws IOException {
    return writeBufferedImage(image, null, null, null);
  }

  protected static byte[] writeBufferedImage(BufferedImage image, CbfMetadata mdata)
      throws IOException {
    return writeBufferedImage(image, mdata, null);
  }

  protected static byte[] writeBufferedImage(BufferedImage image, ImageWriteParam param)
      throws IOException {
    return writeBufferedImage(image, null, param);
  }

  protected static byte[] writeBufferedImage(BufferedImage image, ImageWriteParam param,
      IIOWriteProgressListener listener) throws IOException {
    return writeBufferedImage(image, null, param, listener);
  }

  protected static byte[] writeBufferedImage(BufferedImage image, CbfMetadata mdata,
      ImageWriteParam param) throws IOException {
    return writeBufferedImage(image, mdata, param, null);
  }

  protected static byte[] writeBufferedImage(BufferedImage image, CbfMetadata mdata,
      ImageWriteParam param, IIOWriteProgressListener listener) throws IOException {
    ByteArrayOutputStream outBytesStream = new ByteArrayOutputStream();
    ImageOutputStream outStream = new MemoryCacheImageOutputStream(outBytesStream);
    CbfImageWriterSpi writerSpi = new CbfImageWriterSpi();
    CbfImageWriter writer = (CbfImageWriter)writerSpi.createWriterInstance();
    if (listener != null) writer.addIIOWriteProgressListener(listener);
    writer.setOutput(outStream);
    writer.write(null, new IIOImage(image, null, mdata), param);
    if (listener != null) writer.removeIIOWriteProgressListener(listener);
    return outBytesStream.toByteArray();
  }

  protected static ImageReadParam createImageReadParam() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    return reader.getDefaultReadParam();
  }

  protected static ImageWriteParam createImageWriteParam() throws IOException {
    CbfImageWriterSpi spi = new CbfImageWriterSpi();
    CbfImageWriter writer = (CbfImageWriter)spi.createWriterInstance();
    return writer.getDefaultWriteParam();
  }
}
